/*! \file 
    \brief Matice
*/ 

#ifndef H__MATRIX
#define H__MATRIX

#include <iostream>


using namespace std;

/// Tda matice pouvan pro RBF interpolaci
template <class T> class matrix
{
  public:
	  
    matrix(int height, int width);
    matrix(const matrix &);
    ~matrix();
  
    void operator = (matrix);
	matrix trans(void) const;
	matrix inv(void) const;


    T & operator () (int, int) const;
    class exception
    {
      public:
      const char *message;
      exception(const char *s) {message = s;}
    };
    static matrix I(int);
    
  public:
    int height, width;
    T *cells;
    
};

template <class T> istream & operator >> (istream &in, matrix<T> &A)
{
  for (int i = 0; i < A.height; i++)
  {
    for (int j = 0; j < A.width; j++)
      in >> A(i,j);
  }
  return in;
}

template <class T> ostream & operator << (ostream &out, matrix<T> A)
{
  int width = out.width();
  int precision = out.precision();
  long flags = out.flags();
  for (int i = 0; i < A.height; i++)
  {
    for (int j = 0; j < A.width; j++)
    {
      out.width(width);
      out.precision(precision);
      out.flags(flags);
      out << A(i,j);
    }
    out << endl;
  }
  return out;
}

template <class T> matrix<T>::matrix(int height, int width)
{
  this->height = height;
  this->width = width;
  int n = height*width;
  cells = new T[n];
}

template <class T> matrix<T>::matrix(const matrix<T> &A)
{
  height = A.height;
  width = A.width;
  cells = new T[height*width];
  memcpy(cells, A.cells, height*width*sizeof(T));
}

template <class T> void matrix<T>::operator = (matrix<T> A)
{
  if (height != A.height || width != A.width)
    throw matrix<T>::exception("Incompatible matrix dimensions");
  height = A.height;
  width = A.width;
  memcpy(cells, A.cells, height*width*sizeof(T));
}

template <class T> matrix<T>::~matrix()
{
  delete cells;
}

template <class T> T &matrix<T>::operator () (int row, int column) const
{
  return cells[(row)*width + column];
}

template <class T> matrix<T> operator + (matrix<T> A, matrix<T> B)
{
  if (A.height != B.height || A.width != B.width)
    throw matrix<T>::exception("Incompatible matrix dimensions");
  int i;
  for (i = 0; i < A.height; i++)
  {
    int j;
    for (j = 0; j < A.width; j++)
      A(i,j) = A(i,j) + B(i,j);
  }
  return A;
}

template <class T> bool operator == (matrix<T> A, matrix<T> B)
{
  if (A.height != B.height || A.width != B.width)
    throw matrix<T>::exception("Incompatible matrix dimensions");
  int i;
  for (i = 0; i < A.height; i++)
  {
    int j;
    for (j = 0; j < A.width; j++)
    {
      if (!(A(i,j) == B(i,j)))
        return false;
    }
  }
  return true;
}

template <class T> matrix<T> operator - (matrix<T> A)
{
  int i;
  for (i = 0; i < A.height; i++)
  {
    int j;
    for (j = 0; j < A.width; j++)
      A(i,j) = -A(i,j);
  }
  return A;
}

template <class T> matrix<T> operator * (T x, matrix<T> A)
{
  int i;
  for (i = 0; i < A.height; i++)
  {
    int j;
    for (j = 0; j < A.width; j++)
      A(i,j) = x * A(i,j);
  }
  return A;
}

template <class T> matrix<T> operator - (matrix<T> A, matrix<T> B)
{
  if (A.height != B.height || A.width != B.width)
    throw matrix<T>::exception("Incompatible matrix dimensions");
  int i;
  for (i = 0; i < A.height; i++)
  {
    int j;
    for (j = 0; j < A.width; j++)
      A(i,j) = A(i,j) - B(i,j);
  }
  return A;
}

template <class T> matrix<T> operator * (matrix<T> A, matrix<T> B)
{
  if (A.width != B.height)
    throw matrix<T>::exception("Incompatible matrix dimensions");
  int i;
  matrix<T> AB(A.height, B.width);
  for (i = 0; i < AB.height; i++)
  {
    int j;
    for (j = 0; j < AB.width; j++)
    {
      AB(i,j) = 0.0;
      int k;
      for (k = 0; k < A.width; k++)
        AB(i,j) = AB(i,j) + A(i,k) * B(k,j);
    }
  }
  return AB;
}

template <class T> matrix<T> matrix<T>::trans(void) const
{
  matrix<T> T(width, height);
  int i;
  for (i = 0; i < T.height; i++)
  {
    int j;
    for (j = 0; j < T.width; j++)
      T(i,j) = (*this)(j,i);
  }
  return T;
}

template <class T> matrix<T> matrix<T>::I(int n)
{
  if (n < 1) n = 1;
  matrix<T> U(n, n);

  for (int i = 0; i < n; i++)
	for (int j = 0; j < n; j++)
		if ( i == j ) U(i,j) = 1.0f;
		else U(i,j) = 0.0f;
  return U;
}

template <class T> matrix<T> matrix<T>::inv(void) const
{
  if (height != width)
    throw exception("Inversion of non-square matrix");
  const int N = height;
  matrix<T> B(*this), R(N,N);
  R = I(N);

  for (int k = 0; k < N; k++)
  {
    int pivot_row = -1;
    T pivot_value;
    T abs_pivot_value;

	// vyber pivota, nejvetsi prvek v absolutni hodnote
    for (int i = k; i < N; i++)
    {
		T x = B(i,k);
		T absx = x;
		if (absx < 0) absx = -absx;
		if (pivot_row == -1 || absx > abs_pivot_value)
		{
	        pivot_row = i;
		    abs_pivot_value = absx;
			pivot_value = x;
		}
		if (pivot_value == 0.0) 
			throw exception("Inversion of singular matrix");
    }

    if (pivot_row == k)
    {
      for (int j = k; j < N; j++)
        B(k,j) = B(k,j) / pivot_value;
      for (int j = 0; j < N; j++)
        R(k,j) = R(k,j) / pivot_value;
    }
    else
    {
      for (int j = k; j < N; j++)
      {
        T x = B(k,j);
        B(k,j) = B(pivot_row,j) / pivot_value;
        B(pivot_row,j) = x;
      }
      for (int j = 0; j < N; j++)
      {
        T x = R(k,j);
        R(k,j) = R(pivot_row,j) / pivot_value;
        R(pivot_row,j) = x;
      }
    }
    for (int i = 0; i < N; i++)
    {
      if (i != k)
      {
        for (int j = k+1; j < N; j++)
          B(i,j) = B(i,j) - B(i,k) * B(k,j);
        for (int j = 0; j < N; j++)
          R(i,j) = R(i,j) - B(i,k) * R(k,j);
      }
    }
  }
  return R;
}

#endif
