#include "dxstdafx.h"


#include "CMesh.h"
#include "CSkinningMethod.h"
#include "CAllocateHierarchy.h"
#include "VertexDecl.h"
#include "CMeshMenu.h"

#include <assert.h>


CMesh::CMesh(std::wstring filename, bool Weld, bool Optimize) : m_strFilename(filename)
{
	m_pSM = NULL;
	m_pBoneMatrices = NULL;
	m_pAnimController = NULL;
	m_pFrameRoot = NULL;
	m_pFrameParent = NULL;
	m_pMeshContainer = NULL;
	m_pFrameSelected = NULL;
	m_bWeld = Weld;
	m_bOptimize = Optimize;
	m_ErrMessage = L"";

	m_pMeshMenu = new CMeshMenu(this);
}


CMesh::~CMesh()
{
}


ID3DXAllocateHierarchy* CMesh::GetNewAlloc()
{
	return new CAllocateHierarchy(this);
}


HRESULT CMesh::OnCreateDevice(IDirect3DDevice9* pd3dDevice, DWORD dwBehaviorFlags)
{
	HRESULT hr;

	ID3DXAllocateHierarchy* Alloc = GetNewAlloc();

	V_RETURN( D3DXLoadMeshHierarchyFromX( m_strFilename.c_str(), D3DXMESH_MANAGED, pd3dDevice,
		Alloc, NULL, &m_pFrameRoot, &m_pAnimController ) );
	V_RETURN( SetupBonePointers( m_pFrameRoot ) );
	V_RETURN( D3DXFrameCalculateBoundingSphere( m_pFrameRoot, &m_vObjectCenter, &m_fObjectRadius ) );

	AdvanceTime(0.0f);

	AfterLoad(pd3dDevice);

	if (m_bOptimize) Optimize(pd3dDevice);

	delete Alloc;

	m_dwBehaviorFlags = dwBehaviorFlags;

	return S_OK;
}




HRESULT CMesh::Optimize(IDirect3DDevice9 *pd3dDevice)
{
	D3DXWELDEPSILONS Epsilons;
	HRESULT hr;

	Epsilons.Position	= 0.001f;
	Epsilons.Normal		= 0.001f;


	if (m_bWeld)	
	{
	
		V_RETURN(D3DXWeldVertices(m_pMeshContainer->MeshData.pMesh, D3DXWELDEPSILONS_WELDPARTIALMATCHES, &Epsilons, 
								m_pMeshContainer->pAdjacency, m_pMeshContainer->pAdjacency, NULL, NULL));
	}

	V_RETURN(m_pMeshContainer->MeshData.pMesh->OptimizeInplace(/*D3DXMESH_MANAGED|*/D3DXMESHOPT_VERTEXCACHE,m_pMeshContainer->pAdjacency,
					m_pMeshContainer->pAdjacency, NULL, NULL) );

	return S_OK;
}






void CMesh::OnDestroyDevice()
{
	ID3DXAllocateHierarchy* Alloc = GetNewAlloc();


	if (m_pFrameRoot)
	{
		D3DXFrameDestroy( m_pFrameRoot, Alloc );
		m_pFrameRoot = NULL;
	}

	SAFE_RELEASE( m_pAnimController );

	m_pMeshContainer = NULL;

	delete Alloc;
}



void CMesh::AllocForBones(int NumBones)
{

	m_pBoneMatrices  = new D3DXMATRIX[NumBones];
	m_NumBones = NumBones;
}


void CMesh::FreeForBones()
{
	delete[] m_pBoneMatrices; 
	m_pBoneMatrices = NULL;
}




void CMesh::GetFrameNames(std::vector<std::wstring> &FrameNames)
{
	FrameNames.clear();
	GetFrameNamesRec(m_pFrameRoot, FrameNames);
}

void CMesh::GetFrameNamesRec(LPD3DXFRAME pFrame, std::vector<std::wstring> &FrameNames)
{
	if (pFrame == NULL) return;
	if (pFrame->Name)
		if (strcmp("", pFrame->Name))
		{
			WCHAR strpom[255];
			wsprintf(strpom,L"%S",pFrame->Name);
			std::wstring wsFrameName(strpom);
			FrameNames.push_back(wsFrameName);
		}


	GetFrameNamesRec(pFrame->pFrameSibling, FrameNames);
	GetFrameNamesRec(pFrame->pFrameFirstChild, FrameNames);
}


void CMesh::SelectFrame(LPCSTR BoneName)
{
	m_pFrameSelected = D3DXFrameFind(m_pFrameRoot, BoneName);
	if (m_pFrameSelected == NULL) return;

	m_matTMFrameSelected = ((D3DXFRAME_DERIVED*)m_pFrameSelected)->InitialTransformationMatrix;
}



void CMesh::TransformSelectedFrame(D3DXMATRIX *mat)
{
	if (m_pFrameSelected == NULL) return;

	((D3DXFRAME_DERIVED*)m_pFrameSelected)->TransformationMatrix = (*mat)*m_matTMFrameSelected;

	UpdateFrames(m_pFrameRoot, NULL);
}


void CMesh::SetInitialPose()
{
	SetInitialPoseRec(m_pFrameRoot);
	UpdateFrames(m_pFrameRoot,NULL);
}

void CMesh::SetInitialPoseRec(LPD3DXFRAME pFrameBase)
{
	if (!pFrameBase) return;

	D3DXFRAME_DERIVED* pFrame = (D3DXFRAME_DERIVED*) pFrameBase;

	pFrame->TransformationMatrix = pFrame->InitialTransformationMatrix;

	SetInitialPoseRec(pFrame->pFrameSibling);
	SetInitialPoseRec(pFrame->pFrameFirstChild);
}

//--------------------------------------------------------------------------------------
// Called to setup the pointers for a given bone to its transformation matrix
//--------------------------------------------------------------------------------------
HRESULT CMesh::SetupBonePointers( LPD3DXFRAME pFrame )
{
	HRESULT hr;

	D3DXFRAME_DERIVED* pFrameDerived = (D3DXFRAME_DERIVED*) pFrame;

	pFrameDerived->InitialTransformationMatrix = pFrameDerived->TransformationMatrix;

	if (pFrame->pMeshContainer != NULL)
	{
		hr = SetupBonePointersOnMesh(pFrame->pMeshContainer, pFrame);
		if (FAILED(hr))
			return hr;
	}

	if (pFrame->pFrameSibling != NULL)
	{
		hr = SetupBonePointers(pFrame->pFrameSibling);
		if (FAILED(hr))
			return hr;
	}

	if (pFrame->pFrameFirstChild != NULL)
	{
		hr = SetupBonePointers(pFrame->pFrameFirstChild);
		if (FAILED(hr))
			return hr;
	}

	return S_OK;
}


//--------------------------------------------------------------------------------------
// Called to setup the pointers for a given bone to its transformation matrix
//--------------------------------------------------------------------------------------
void CMesh::SetupBonePointersOnMeshCB(LPD3DXMESHCONTAINER pMeshContainerBase, LPD3DXFRAME pFrameBase, int iBone)
{
	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;
	D3DXFRAME_DERIVED* pFrame = (D3DXFRAME_DERIVED*) pFrameBase;

	pMeshContainer->ppBoneMatrixPtrs[iBone] = &pFrame->CombinedTransformationMatrix;
}


void CMesh::SetupBonePointersOnMeshAlloc(LPD3DXMESHCONTAINER pMeshContainerBase, int iBones)
{
	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;

	pMeshContainer->ppBoneMatrixPtrs = new D3DXMATRIX*[iBones];
}


HRESULT CMesh::SetupBonePointersOnMesh( LPD3DXMESHCONTAINER pMeshContainerBase, LPD3DXFRAME pFrameParent )
{
	UINT iBone, cBones;
	D3DXFRAME_DERIVED *pFrame;

//	if (m_pMeshContainer != NULL) return S_OK/*assert(!"Supported is only one mesh per file")*/;
//	if (pMeshContainerBase->pSkinInfo == NULL) return S_OK;

	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;
	m_pMeshContainer = pMeshContainer;
	m_pFrameParent = (D3DXFRAME_DERIVED*) pFrameParent;

	// if there is a skinmesh, then setup the bone matrices
	if (pMeshContainer->pSkinInfo != NULL)
	{
		cBones = pMeshContainer->pSkinInfo->GetNumBones();

		SetupBonePointersOnMeshAlloc(pMeshContainer, cBones);

		for (iBone = 0; iBone < cBones; iBone++)
		{
			LPCSTR  a = pMeshContainer->pSkinInfo->GetBoneName(iBone);
			pFrame = (D3DXFRAME_DERIVED *) D3DXFrameFind( m_pFrameRoot, pMeshContainer->pSkinInfo->GetBoneName(iBone) );
			if (pFrame == NULL)
				return E_FAIL;

			if(m_pAnimController) m_pAnimController->RegisterAnimationOutput(pFrame->Name,&pFrame->TransformationMatrix,&pFrame->ScaleRel,&pFrame->QuatRel,&pFrame->TransRel);
			SetupBonePointersOnMeshCB(pMeshContainer, pFrame, iBone);
		}
	}

	return S_OK;
}




void CMesh::AdvanceTime(double fElapsedTime)
{
	D3DXMATRIX matId;
	D3DXMatrixIdentity(&matId);

	if( m_pAnimController != NULL )
		m_pAnimController->AdvanceTime( fElapsedTime, NULL );

	UpdateFrames( m_pFrameRoot,NULL);
}


void CMesh::UpdateFramesPreCB(LPD3DXFRAME pFrameBase,LPD3DXFRAME pFrameParent)
{

	D3DXFRAME_DERIVED *pFrame = (D3DXFRAME_DERIVED*)pFrameBase;

	if (pFrameParent == NULL) 
		pFrame->CombinedTransformationMatrix = pFrame->TransformationMatrix;

	else
	{
		D3DXMATRIX &mat = ((D3DXFRAME_DERIVED*) pFrameParent)->CombinedTransformationMatrix;
		D3DXMatrixMultiply(&pFrame->CombinedTransformationMatrix, &pFrame->TransformationMatrix, &mat );
	}
}



void CMesh::UpdateFrames( LPD3DXFRAME pFrameBase, LPD3DXFRAME pFrameParent )
{
	D3DXFRAME_DERIVED *pFrame = (D3DXFRAME_DERIVED*)pFrameBase;

	UpdateFramesPreCB(pFrameBase,pFrameParent);

	if (pFrame->pFrameSibling != NULL)
	{
		UpdateFrames(pFrame->pFrameSibling, pFrameParent);
	}

	if (pFrame->pFrameFirstChild != NULL)
	{
		UpdateFrames(pFrame->pFrameFirstChild, pFrameBase);
	}

	UpdateFramesPostCB(pFrameBase,pFrameParent);

}








void CMesh::Render(IDirect3DDevice9* pd3dDevice)
{
	DrawFrame(pd3dDevice,m_pFrameRoot);
}




//-------------------------------------------------------------------------------------

void CMesh::DrawFrame(IDirect3DDevice9* pd3dDevice, D3DXFRAME *pFrame)
{
	if (!pFrame) return;

    LPD3DXMESHCONTAINER pMeshContainer = pFrame->pMeshContainer;

	while (pMeshContainer != NULL)
	{
		DrawMeshContainer( pd3dDevice, pMeshContainer, pFrame);
		pMeshContainer = pMeshContainer->pNextMeshContainer;
	}

	if (pFrame->pFrameSibling != NULL)
		DrawFrame( pd3dDevice, pFrame->pFrameSibling);


	if (pFrame->pFrameFirstChild != NULL)
		DrawFrame( pd3dDevice, pFrame->pFrameFirstChild);
}





//-------------------------------------------------------------------------------------

void CMesh::SetBonesToEffect( IDirect3DDevice9 *pd3dDevice, UINT iAttrib, LPD3DXMESHCONTAINER pMeshContainerBase)
{
	D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;
	HRESULT hr;
	UINT iMatrixIndex;
	UINT iPaletteEntry;
	LPD3DXBONECOMBINATION pBoneComb;

	pBoneComb = reinterpret_cast<LPD3DXBONECOMBINATION>(pMeshContainer->pBoneCombinationBuf->GetBufferPointer());

	// first calculate all the world matrices
	for (iPaletteEntry = 0; iPaletteEntry < pMeshContainer->NumPaletteEntries; ++iPaletteEntry)
	{
		iMatrixIndex = pBoneComb[iAttrib].BoneId[iPaletteEntry];
		if (iMatrixIndex != UINT_MAX)
		{
			D3DXMatrixMultiply(&m_pBoneMatrices[iPaletteEntry],&pMeshContainer->pBoneOffsetMatrices[iMatrixIndex], pMeshContainer->ppBoneMatrixPtrs[iMatrixIndex]);
		}
	}
	V( m_pSM->GetD3DXEffect()->SetMatrixArray( "mWorldMatrixArray", m_pBoneMatrices, pMeshContainer->NumPaletteEntries) );
}



void CMesh::DrawMeshContainer( IDirect3DDevice9 *pd3dDevice, LPD3DXMESHCONTAINER pMeshContainerBase, LPD3DXFRAME pFrameBase)
{
    D3DXMESHCONTAINER_DERIVED *pMeshContainer = (D3DXMESHCONTAINER_DERIVED*)pMeshContainerBase;
    D3DXFRAME_DERIVED *pFrame = (D3DXFRAME_DERIVED*)pFrameBase;
	HRESULT hr;
    UINT iAttrib;
    LPD3DXBONECOMBINATION pBoneComb;

	if (pMeshContainer->pSkinInfo == NULL) 
	{

		// render bez vertex shaderu
//		D3DXMatrixIdentity(&pFrame->CombinedTransformationMatrix);
		V( pd3dDevice->SetTransform(D3DTS_WORLD, &pFrame->CombinedTransformationMatrix) );

		for (int iMaterial = 0; iMaterial < pMeshContainer->NumMaterials; iMaterial++)
		{
			V( pd3dDevice->SetMaterial( &pMeshContainer->pMaterials[iMaterial].MatD3D ) );
			V( pd3dDevice->SetTexture( 0, pMeshContainer->ppTextures[iMaterial] ) );
			V( pMeshContainer->MeshData.pMesh->DrawSubset(iMaterial) );
		}
		return;
	}




	pBoneComb = reinterpret_cast<LPD3DXBONECOMBINATION>(pMeshContainer->pBoneCombinationBuf->GetBufferPointer());
	for (iAttrib = 0; iAttrib < pMeshContainer->NumAttributeGroups; iAttrib++)
	{ 

		SetBonesToEffect(pd3dDevice, iAttrib,pMeshContainer);

		// Sum of all ambient and emissive contribution
		D3DXCOLOR color1(pMeshContainer->pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Ambient);
		D3DXCOLOR color2(.25, .25, .25, 1.0);
		D3DXCOLOR ambEmm;
		D3DXColorModulate(&ambEmm, &color1, &color2);
		ambEmm += D3DXCOLOR(pMeshContainer->pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Emissive);

		// set material color properties 
		V( m_pSM->GetD3DXEffect()->SetVector("MaterialDiffuse", (D3DXVECTOR4*)&(pMeshContainer->pMaterials[pBoneComb[iAttrib].AttribId].MatD3D.Diffuse)) );
		V( m_pSM->GetD3DXEffect()->SetVector("MaterialAmbient", (D3DXVECTOR4*)&ambEmm) );

		// setup the material of the mesh subset - REMEMBER to use the original pre-skinning attribute id to get the correct material id
		V( pd3dDevice->SetTexture( 0, pMeshContainer->ppTextures[pBoneComb[iAttrib].AttribId] ) );

		// Set CurNumBones to select the correct vertex shader for the number of bones
		V( m_pSM->GetD3DXEffect()->SetInt( "CurNumBones", pMeshContainer->NumInfl -1) );

		// Start the effect now all parameters have been updated
		UINT numPasses;
		V( m_pSM->GetD3DXEffect()->Begin( &numPasses, D3DXFX_DONOTSAVESTATE ) );
		for( UINT iPass = 0; iPass < numPasses; iPass++ )
		{
			V( m_pSM->GetD3DXEffect()->BeginPass( iPass ) );

			// draw the subset with the current world matrix palette and material state
			V( pMeshContainer->MeshData.pMesh->DrawSubset( iAttrib ) );

			V( m_pSM->GetD3DXEffect()->EndPass() );
		}

		V( m_pSM->GetD3DXEffect()->End() );

		V( pd3dDevice->SetVertexShader(NULL) );
	}

}



void	CMesh::ProcessVertices(VERTEX **ppVertices)
{
	if (m_pMeshContainer->pSkinInfo )
		ProcessVerticesSkinned(ppVertices);
	else
		ProcessVerticesRigid(ppVertices);

}


void	CMesh::ProcessVerticesSkinned(VERTEX **ppVertices)
{
}



void CMesh::ProcessVerticesRigid(VERTEX **ppVertices)
{
	VERTEX* pData;
	LPD3DXMESH pMesh = m_pMeshContainer->MeshData.pMesh;
	DWORD dwVertices = pMesh->GetNumVertices();
	DWORD dwStride = pMesh->GetNumBytesPerVertex();

	D3DXMATRIX &mat = ((D3DXFRAME_DERIVED*)m_pFrameParent)->CombinedTransformationMatrix;

	*ppVertices = new VERTEX[dwVertices];

	if (pMesh->LockVertexBuffer(0,(void**)&pData) == D3D_OK)
	{
		for(DWORD i=0; i<dwVertices; ++i)
		{

			D3DXVec3TransformCoord(&(*ppVertices)[i].position,&pData->position,&mat);
			D3DXVec3TransformNormal(&(*ppVertices)[i].normal,&pData->normal,&mat);

			pData = (VERTEX*)(((BYTE*)pData) + dwStride);
		}

		pMesh->UnlockVertexBuffer();
	}

}




LPD3DXFRAME CMesh::GetFrameByName(LPCSTR BoneName)
{
	return D3DXFrameFind(m_pFrameRoot, BoneName);
}


